import java.io.File;
import java.io.FileNotFoundException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Scanner;
import java.util.stream.Collectors;

/**
 * An illustration of how one might augment the raw CSV solvability data
 * with normalized features to support learning classification of 
 * solvable versus unsolvable states.  There are many other features of
 * potential interest, but I leave that to your creativity and curiosity.
 * 
 * I would strongly encourage you to start out as follows:
 * - make a list of potential features that (1) would be computationally
 *   simple, and (2) may correlate strongly with solvability 
 * - rewrite this code so as to make a dataset with richer features
 * - gather resources for learning more about classification: 
 * -- https://sigai.acm.org/static/aimatters/3-2/AIMatters-3-2-05-Neller.pdf
 * -- http://cs.gettysburg.edu/ai-matters/index.php/Resources#Machine_Learning 
 * - learn about one or two tools to support such learning, e.g.
 * -- Weka (https://www.cs.waikato.ac.nz/ml/weka/)
 * -- R Studio (https://www.rstudio.com/) with this text (http://www-bcf.usc.edu/~gareth/ISL/)
 * - start out with simple linear regression on normalized features to see which matter most
 * - use such insight for node ordering in simple depth-first search
 * - measure search performance using good empirical research practice
 * - continue adding onto your skill set with slow, regular progress
 * 
 * @author Todd W. Neller
 *
 */
public class DataFeatureAdder {
	public static final int STATE_IDX = 0, CARDS_IDX = 1, CHILDREN_IDX = 2, RATIO_IDX = 3, SOLVABLE_IDX = 4; // layout of new CSV records
	public String filename; // input filename (without .csv extension)
	public ArrayList<Object[]> dataWithFeatures = new ArrayList<>(); // new augmented data
	// Note each Object[] will contain state string, normalized cards, 
	// normalized number of children, normalized ratio of children to cards, 
	// and 1/0 solvability.
	
	/**
	 * Read in comma-separated value (CSV) data from the given input filename, add new features for classification,
	 * and normalize such features to values ranging from 0 to 1.
	 * @param filename given input filename (without .csv extension)
	 */
	public DataFeatureAdder(String filename) {
		this.filename = filename;
		readCSVSolvabilityData();
		addFeatures(); 
		normalizeFeatures();
	}
	
	/**
	 * Read the CSV data generated by SolvabilityDataGenerator.java.
	 */
	private void readCSVSolvabilityData() {
		try (Scanner in = new Scanner(new File(filename + ".csv"));) {
			in.nextLine(); // skip "state","solvable" line 
			while (in.hasNextLine()) {
				String line = in.nextLine();
				String[] data = line.split(",");
				Object[] dataPlus = new Object[SOLVABLE_IDX + 1];
				dataPlus[STATE_IDX] = data[0];
				dataPlus[SOLVABLE_IDX] = data[1];
				dataWithFeatures.add(dataPlus);
			}
		}
		catch (Exception e) {
			e.printStackTrace();
			System.exit(1);
		}
	}	
	
	/**
	 * Augment the data with your chosen features potentially helpful to classification.
	 */
	private void addFeatures() { 
		for (Object[] entry : dataWithFeatures) {
			// reconstruct node
			BirdsOfAFeatherNode node = new BirdsOfAFeatherNode();
			int index = 1;
			int numCards = 0;
			for (int r = 0; r < 4; r++)
				for (int c = 0; c < 4; c++) {
					String cardStr = ((String) entry[0]).substring(index, index + 2);
					node.grid[r][c] = cardStr.equals("--") ? null : Card.strCardMap.get(cardStr);
					if (node.grid[r][c] != null)
						numCards++;
					index += 2;
				}
			entry[CARDS_IDX] = numCards;
			entry[CHILDREN_IDX] = node.expand().size();
			entry[RATIO_IDX] = ((Integer) entry[2]).doubleValue() / (Integer) entry[1];
		}
	}

	/**
	 * Normalize the features we're adding.
	 */
	private void normalizeFeatures() {
		normalizeFeature(CARDS_IDX);
		normalizeFeature(CHILDREN_IDX);
		normalizeFeature(RATIO_IDX);		
	}

	/**
	 * Scale and translate values to 0.0 (minimum) through 1.0 (maximum).
	 * @param index index of data to normalize
	 */
	private void normalizeFeature(int index) {
		double min = Double.MAX_VALUE;
		double max = Double.MIN_VALUE;
		for (Object[] entry : dataWithFeatures) {
			double value = ((Number) entry[index]).doubleValue();
			if (value < min)
				min = value;
			if (value > max)
				max = value;
		}
		double range = max - min;
		if (range == 0)
			return;
		for (Object[] entry : dataWithFeatures) {
			double value = ((Number) entry[index]).doubleValue();
			entry[index] = (value - min) / range;
		}
	}

	/**
	 * Output new data CSV file augmented with our new features.
	 * @param outFilename name of output CSV file
	 */
	public void exportCSV(String outFilename) {
		try (PrintWriter out = new PrintWriter(outFilename + ".csv");) {
			out.println("\"state\",\"num_cards\",\"num_moves\",\"ratio_moves_per_card\",\"solvable\"");
			for (Object[] entry : dataWithFeatures)
				out.println(Arrays.asList(entry).stream().map(a -> a.toString()).collect(Collectors.joining(",")));
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		}
	}

	/**
	 * Take the raw solvability data output from SolvabilityDataGenerator.java
	 * and add new features to support solvability classification.
	 * @param args (unused)
	 */
	public static void main(String[] args) {
		new DataFeatureAdder("boaf-data-0-100").exportCSV("boaf-data-0-100-with-features");
	}

}
