import java.io.File;
import java.io.FileNotFoundException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Random;

public class SolvabilityDataGenerator {
	public String filename; // output filename (without .csv extension)
	public int startSeed; // beginning seed for data generation
	public int minEndSeed = Integer.MAX_VALUE; // first seed not included in data generation
	public int minStates = Integer.MAX_VALUE; // minimum number of states output in data generation
	public int minSolvableStates = Integer.MAX_VALUE; // minimum number of solvable states output in data generation
	public int minUnsolvableStates = Integer.MAX_VALUE; // minimum number of unsolvable states output in data generation

	public SolvabilityDataGenerator(String filename, int startSeed, int numSeeds) {
		this.filename = filename;
		this.startSeed = startSeed;
		minEndSeed = startSeed + numSeeds;
	}

	/**
	 * Generates comma-separated value (CSV) data where each line consists of a 
	 * String description of a Birds of a Feather state, a comma, and an integer 
	 * 0/1 indicating that the state is unsolvable/solvable.  The initial state 
	 * of a seed is always included in the data. For each solvable initial 
	 * state, the following iterative process is used to generate data through 
	 * simulated play:
	 * <ul>
	 * <li>All children of the state are generated.</li>
	 * <li>If there are no solvable children, we increment the seed.</li>
	 * <li>If all are solvable, pick one at random as the next state, generate no
	 *     output, and repeat.</li>
	 * <li>If some are unsolvable, print a CSV line for each child, pick a
	 *     solvable child at random as the next state, and repeat.</li>
	 * </ul>
	 * This process continues until we've met or exceeded one minimum stopping 
	 * condition as expressed in our public fields (e.g. minEndSeed, minStates,
	 * minSolvableStates, minUnsolvableStates). States are counted if they are
	 * output.  We do not include children that are all solvable on the output,
	 * as we wish to focus our machine learning on state decisions where the
	 * decision matters.
	 * @throws FileNotFoundException 
	 */
	public void generateCSVData() throws FileNotFoundException {
		Random random = new Random(0); // for reproducability
		int seed = startSeed;
		int stateCount = 0;
		int solvableStateCount = 0;
		int unsolvableStateCount = 0;
		File file = new File(filename + ".csv");
		PrintWriter out = new PrintWriter(file);
		out.println("\"state\",\"solvable\"");
		while (seed < minEndSeed && stateCount < minStates && solvableStateCount < minSolvableStates && unsolvableStateCount < minUnsolvableStates) {
			BirdsOfAFeatherNode state = new BirdsOfAFeatherNode(seed++);
			SolvabilitySearcher searcher = new SolvabilitySearcher();
			boolean solvable = searcher.search(state);
			out.printf("\"%s\",%d\n", stateToCSVString(state), solvable ? 1 : 0);
			System.out.printf("\"%s\",%d\n", stateToCSVString(state), solvable ? 1 : 0);
			if (!solvable)
				continue;
			ArrayList<BirdsOfAFeatherNode> solvableChildren = new ArrayList<BirdsOfAFeatherNode>();
			ArrayList<BirdsOfAFeatherNode> unsolvableChildren = new ArrayList<BirdsOfAFeatherNode>();
			while (true) {
				ArrayList<SearchNode> children = state.expand();
				if (children.isEmpty())
					break;
				stateCount += children.size();
				solvableChildren.clear();
				unsolvableChildren.clear();
				for (SearchNode child : children)
					(searcher.search(child) ? solvableChildren : unsolvableChildren).add((BirdsOfAFeatherNode) child);
				state = solvableChildren.get(random.nextInt(solvableChildren.size()));
				if (unsolvableChildren.isEmpty())
					continue;
				solvableStateCount += solvableChildren.size();
				unsolvableStateCount += unsolvableChildren.size();
				for (BirdsOfAFeatherNode child : solvableChildren)
					out.printf("\"%s\",1\n", stateToCSVString(child));
				for (BirdsOfAFeatherNode child : unsolvableChildren)
					out.printf("\"%s\",0\n", stateToCSVString(child));
			}
		}
		out.close();
		
	}
	
	/**
	 * Convert the puzzle state of the given BirdsOfAFeatherNode to a row-major ordered list of
	 * two-character codes Card toString() or "--" if a grid cell is empty.
	 * @param state given BirdsOfAFeatherNode
	 * @return a row-major ordered list of two-character codes describing the state
	 */
	public String stateToCSVString(BirdsOfAFeatherNode state) {
		StringBuilder sb = new StringBuilder();
		for (int r = 0; r < state.grid.length; r++)
			for (int c = 0; c < state.grid[0].length; c++) {
				Card card = state.grid[r][c];
				sb.append(card == null ? "--" : card.toString());
			}
		return sb.toString();
	}

	/**
	 * Generate a comma-separated value (CSV) file of Birds of a Feather states along with whether
	 * each state is solvable (1), or unsolvable (0).
	 * @param args (unused)
	 * @throws FileNotFoundException
	 */
	public static void main(String[] args) throws FileNotFoundException {
		SolvabilityDataGenerator generator = new SolvabilityDataGenerator("boaf-data-0-100", 0, 100);
//		SolvabilityDataGenerator generator = new SolvabilityDataGenerator("boaf-data-100-1000", 100, 1000);
//		SolvabilityDataGenerator generator = new SolvabilityDataGenerator("boaf-data-1100-10000", 1100, 10000);
		// can also request other stopping conditions, e.g. minimum number of solvable states:
		// generator.minSolvableStates = 1000000;
		generator.generateCSVData();  
	}

}
